!
!  MinionBrownACM.f95 
!
!	Version 1.0 (17 April 2014)
!
!  Acknowledgements to Martin Geier for pointing out a bug, which has been fixed now
!
!  SUBS/FUNCTIONS:
!	MinionBrownACM 			            - Entry point of console application.
!	dx,dy,dxx,dyy,dxy,dxxy,dxyy,dxxyy   - FD functions
!
!****************************************************************************
!
!  PROGRAM: MinionBrownACM
!
!  PURPOSE: ACM scheme
!
!  AUTHORS: Pietro and Taku, Japanese Noodle Fan Club, Torino and Kyoto 
!
!****************************************************************************

program MinionBrownACM

	implicit none

!============================================================================
! Input Data
!============================================================================

! Input Parameters
	real(kind=2)    :: Tt =  1.0d0
	real(kind=2)    :: Lx =  1.0d0
	real(kind=2)    :: UL =  1.0d0
	real(kind=2)    :: P0 =  0.0d0
! kinematic viscosity
	real(kind=2)    :: NU =  0.0001d0 ! = 1/Re
! space discretization
	integer         :: Nx = 128

	real(kind=2)    :: bet = (0.2d0*10)**2
	real(kind=2)    :: gam = 0.d0
	real(kind=2)    :: mu  = 1.d0
	
! Auxiliaries	
	real(kind=2),dimension(:,:),allocatable     :: u,v,P,uh,vh,Ph,vort,div
	real(kind=2),dimension(:),allocatable       :: maxvort
	real(kind=2),dimension(1:3,1:3)             :: us,vs,uus,vvs,uvs,Ps
	real(kind=2)    :: eps,dt,usP,vsP,PsP,u0,v0,P0,x,y,temp,Ly
	real(kind=2)    :: houx,hovy
	real(kind=2)    :: pi = 3.141592653589793238d0 
	integer         :: Ny,Nt,t,i,j,ik,jk
	integer         :: IO_Stat_Number1 = -1
	
! Internal Functions
	real(kind=2)    :: dx,dy,dxx,dyy,dxy,dxxy,dxyy,dxxyy
	
! Initialization
    Ly  = Lx;
    eps = Lx/Nx;
    Ny  = nint(Ly/eps);
    dt  = 2.d0*eps**2;
    Nt  = int(Tt/dt);

	allocate( u(1:(Nx+2),1:(Ny+2)) )    
	allocate( v(1:(Nx+2),1:(Ny+2)) )    
	allocate( P(1:(Nx+2),1:(Ny+2)) )    
	allocate( uh(1:(Nx+2),1:(Ny+2)) )    
	allocate( vh(1:(Nx+2),1:(Ny+2)) )    
	allocate( Ph(1:(Nx+2),1:(Ny+2)) )
	allocate( vort(1:(Nx+2),1:(Ny+2)) )
	allocate( div(1:(Nx+2),1:(Ny+2)) )
	allocate( maxvort(1:Nt) )
	
! Initialization (by equilibrium)
	do i = 1,Nx+2,1
		do j = 1,Ny+2,1
			x = real(i-1)/real(Nx)-1.0d0/(2.0d0*real(Nx));
			y = real(j-1)/real(Ny)-1.0d0/(2.0d0*real(Ny));
		
			if (y<=0.5d0) then
			    u(i,j)  = tanh(80.d0*(y-0.25d0));
			else
			    u(i,j)  = tanh(80.d0*(0.75d0-y));
            endif
			v(i,j)  = 0.05d0*sin(2.d0*pi*(x+0.25d0));
			P(i,j) = P0;
		enddo
	enddo
	uh(:,:) = 0.d0;   
	vh(:,:) = 0.d0;   
	Ph(:,:) = 0.d0;   
    
	do t = 1,Nt,1
	    write(*,*) t,Nt,maxval(u)
	    
!    % Apply boundary conditions to u,v,P
        u(Nx+2,:) = u(2,:); ! E from W
        v(Nx+2,:) = v(2,:);
        P(Nx+2,:) = P(2,:);

        u(1,:) = u(Nx+1,:); ! W from E
        v(1,:) = v(Nx+1,:);
        P(1,:) = P(Nx+1,:);
        
        u(:,Ny+2) = u(:,2); ! N from S
        v(:,Ny+2) = v(:,2);
        P(:,Ny+2) = P(:,2);

        u(:,1) = u(:,Nx+1); ! S from N
        v(:,1) = v(:,Nx+1);
        P(:,1) = P(:,Nx+1);
        
        u(Nx+2,Ny+2) = u(2,2); ! NE from SW
        v(Nx+2,Ny+2) = v(2,2);
        P(Nx+2,Ny+2) = P(2,2);
        
        u(1,Ny+2) = u(Nx+1,2); ! NW from SE
        v(1,Ny+2) = v(Nx+1,2);
        P(1,Ny+2) = P(Nx+1,2);
        
        u(1,1) = u(Nx+1,Ny+1); ! SW from NE
        v(1,1) = v(Nx+1,Ny+1);
        P(1,1) = P(Nx+1,Ny+1);

        u(Nx+2,1) = u(2,Nx+1); ! SE from NW
        v(Nx+2,1) = v(2,Nx+1);
        P(Nx+2,1) = P(2,Nx+1);
        
!    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!    % First half step: From "t" to "t+dt/2"
!    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        do i = 2,(Nx+1),1
            do j = 2,(Ny+1),1
            
                !% Load stencils
	            do ik = 1,3,1
	                do jk = 1,3,1
	                    us(ik,jk) = u(ik+i-2,jk+j-2);
	                    vs(ik,jk) = v(ik+i-2,jk+j-2);
	                    Ps(ik,jk) = P(ik+i-2,jk+j-2);
	                enddo
	            enddo
                usP = us(2,2);
                vsP = vs(2,2);
                PsP = Ps(2,2);
                u0 = u(i,j);
                v0 = v(i,j);
                P0 = P(i,j);
                
                !% time "t+dt/2"
                uh(i,j) = u0+(dt/2.d0)*( -usP*dx(us,eps)-vsP*dy(us,eps)-dx(Ps,eps)+nu*(dxx(us,eps)+dyy(us,eps)) );
                vh(i,j) = v0+(dt/2.d0)*( -usP*dx(vs,eps)-vsP*dy(vs,eps)-dy(Ps,eps)+nu*(dxx(vs,eps)+dyy(vs,eps)) );
            enddo
        enddo
    
!    % Apply boundary conditions to uh,vh
        uh(Nx+2,:) = uh(2,:); ! E from W
        vh(Nx+2,:) = vh(2,:);

        uh(1,:) = uh(Nx+1,:); ! W from E
        vh(1,:) = vh(Nx+1,:);
        
        uh(:,Ny+2) = uh(:,2); ! N from S
        vh(:,Ny+2) = vh(:,2);

        uh(:,1) = uh(:,Nx+1); ! S from N
        vh(:,1) = vh(:,Nx+1);
        
        uh(Nx+2,Ny+2) = uh(2,2); ! NE from SW
        vh(Nx+2,Ny+2) = vh(2,2);
        
        uh(1,Ny+2) = uh(Nx+1,2); ! NW from SE
        vh(1,Ny+2) = vh(Nx+1,2);
        
        uh(1,1) = uh(Nx+1,Ny+1); ! SW from NE
        vh(1,1) = vh(Nx+1,Ny+1);

        uh(Nx+2,1) = uh(2,Nx+1); ! SE from NW
        vh(Nx+2,1) = vh(2,Nx+1);
            
        do i = 2,(Nx+1),1;
            do j = 2,(Ny+1),1;
            
                !% Load stencils
	            do ik = 1,3,1
	                do jk = 1,3,1
                        us(ik,jk) = (1.d0+P(ik+i-2,jk+j-2)*bet*eps**2)*uh(ik+i-2,jk+j-2);
                        vs(ik,jk) = (1.d0+P(ik+i-2,jk+j-2)*bet*eps**2)*vh(ik+i-2,jk+j-2);
                    enddo
                enddo    
                P0 = P(i,j);
                
                !% time "t+dt/2"
                div(i,j) = dx(us,eps)+dy(vs,eps)+(eps**2/6.d0)*(dxxy(vs,eps)+dxyy(us,eps));
                Ph(i,j) = ( P0-(dt/2.d0)*div(i,j)/(bet*eps**2) )/(1+gam*(dt/2.d0));
            enddo
        enddo
        
!    % Apply boundary conditions to Ph
        Ph(Nx+2,:) = Ph(2,:);
        Ph(1,:) = Ph(Nx+1,:);
        Ph(:,Ny+2) = Ph(:,2);
        Ph(:,1) = Ph(:,Nx+1);
        
        Ph(Nx+2,Ny+2) = Ph(2,2);
        Ph(1,Ny+2) = Ph(Nx+1,2);
        Ph(1,1) = Ph(Nx+1,Ny+1);
        Ph(Nx+2,1) = Ph(2,Nx+1);
    
!    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!    % Second half step: From "t+dt/2" to "t+dt"
!    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        do i = 2,(Nx+1),1;
            do j = 2,(Ny+1),1;
            
                !% Load stencils
	            do ik = 1,3,1
	                do jk = 1,3,1                
                        us(ik,jk) = uh(ik+i-2,jk+j-2);
                        vs(ik,jk) = vh(ik+i-2,jk+j-2);
                        Ps(ik,jk) = Ph(ik+i-2,jk+j-2);
                    enddo
                enddo
                usP = us(2,2);
                vsP = vs(2,2);
                PsP = Ps(2,2);
                u0 = u(i,j);
                v0 = v(i,j);
                P0 = P(i,j);

                !% time "t+dt"
                u(i,j) = u0+dt*( -usP*dx(us,eps)-vsP*dy(us,eps)-dx(Ps,eps)+nu*(dxx(us,eps)+dyy(us,eps)) );
                v(i,j) = v0+dt*( -usP*dx(vs,eps)-vsP*dy(vs,eps)-dy(Ps,eps)+nu*(dxx(vs,eps)+dyy(vs,eps)) );

                P(i,j) = P0+dt*( -gam*PsP-div(i,j)/(bet*eps**2)+ &
                    & mu*eps*( dxx(Ps,eps)+dyy(Ps,eps)+2.d0*(dx(vs,eps)*dy(us,eps)-dx(us,eps)*dy(vs,eps)) ) );
            enddo
        enddo
        
!    % Apply boundary conditions to u,v,P
        u(Nx+2,:) = u(2,:); ! E from W
        v(Nx+2,:) = v(2,:);
        P(Nx+2,:) = P(2,:);

        u(1,:) = u(Nx+1,:); ! W from E
        v(1,:) = v(Nx+1,:);
        P(1,:) = P(Nx+1,:);
        
        u(:,Ny+2) = u(:,2); ! N from S
        v(:,Ny+2) = v(:,2);
        P(:,Ny+2) = P(:,2);

        u(:,1) = u(:,Nx+1); ! S from N
        v(:,1) = v(:,Nx+1);
        P(:,1) = P(:,Nx+1);
        
        u(Nx+2,Ny+2) = u(2,2); ! NE from SW
        v(Nx+2,Ny+2) = v(2,2);
        P(Nx+2,Ny+2) = P(2,2);
        
        u(1,Ny+2) = u(Nx+1,2); ! NW from SE
        v(1,Ny+2) = v(Nx+1,2);
        P(1,Ny+2) = P(Nx+1,2);
        
        u(1,1) = u(Nx+1,Ny+1); ! SW from NE
        v(1,1) = v(Nx+1,Ny+1);
        P(1,1) = P(Nx+1,Ny+1);

        u(Nx+2,1) = u(2,Nx+1); ! SE from NW
        v(Nx+2,1) = v(2,Nx+1);
        P(Nx+2,1) = P(2,Nx+1);
    
!    % Calculation of vorticity
        do i = 2,(Nx+1),1
            do j = 2,(Ny+1),1
                    
                !% Load stencils
	            do ik = 1,3,1
	                do jk = 1,3,1
	                    us(ik,jk) = u(ik+i-2,jk+j-2);
	                    vs(ik,jk) = v(ik+i-2,jk+j-2);
	                    Ps(ik,jk) = P(ik+i-2,jk+j-2);
	                enddo
	            enddo
	            
	            vort(i,j) = dx(vs,eps)-dy(us,eps); ! Simple 3-points formula
            enddo
        enddo
        maxvort(t) = maxval(vort);
                    
    enddo
    
! Save
	open(UNIT = 1,FILE = 'results.dat',IOSTAT = IO_Stat_Number1)
	if (IO_Stat_Number1 == 0) then
		print *,'Saving results.dat ...'

		do j = 2,(Ny+1),1
			do i = 2,(Nx+1),1
				x = real(i-1)*Lx/real(Nx)-Lx/(2.0d0*real(Nx));
				y = real(j-1)*Ly/real(Ny)-Ly/(2.0d0*real(Ny));

				write(1,*) x,y,u(i,j)
				write(1,*) v(i,j),P(i,j),vort(i,j)
			enddo
		enddo

		print *, '... done' 
	endif
	close(UNIT=1)    

	open(UNIT = 1,FILE = 'maxvort.dat',IOSTAT = IO_Stat_Number1)
	if (IO_Stat_Number1 == 0) then
		print *,'Saving maxvort.dat ...'

		do t = 1,Nt
            write(1,*) dt*real(t-1),maxvort(t)
		enddo

		print *, '... done' 
	endif
	close(UNIT=1)    
	
! De-allocate    
	deallocate( u )
	deallocate( v )
	deallocate( P )
	deallocate( uh )
	deallocate( vh )
	deallocate( Ph )
	deallocate( vort )
	deallocate( div )
	deallocate( maxvort )
	
end program MinionBrownACM

!&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
!============================================================================
! Utilities
!============================================================================
!&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
function dx(stencil,eps) result(ans)

	implicit none

! Inlet/Outlet
	real(kind=2) :: ans
	real(kind=2),dimension(1:3,1:3),intent(in) :: stencil
	real(kind=2),intent(in) :: eps

    real(kind=2) :: P,E,W,N,S,NE,NW,SE,SW
    
!% NW(1,3) --- N(2,3) --- NE(3,3)
!%
!%  W(1,2) --- P(2,2) ---  E(3,2)
!%
!% SW(1,1) --- S(2,1) --- SE(3,1)
    
    P = stencil(2,2);
    E = stencil(3,2);
    W = stencil(1,2);
    N = stencil(2,3);
    S = stencil(2,1);

    NE = stencil(3,3);
    NW = stencil(1,3);
    SE = stencil(3,1);
    SW = stencil(1,1);
    
    ans = ( E-W )/(2.d0*eps);

!% ! \partial_x \phi 2nd order accuracy
!% 	else if ((part_x==1).and.(part_y==0)) then
!% 		FDoperator = (PhiE-PhiW)/(2.0*dx);

end function dx

!&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
function dy(stencil,eps) result(ans)

	implicit none

! Inlet/Outlet
	real(kind=2) :: ans
	real(kind=2),dimension(1:3,1:3),intent(in) :: stencil
	real(kind=2),intent(in) :: eps

    real(kind=2) :: P,E,W,N,S,NE,NW,SE,SW
    
!% NW(1,3) --- N(2,3) --- NE(3,3)
!%
!%  W(1,2) --- P(2,2) ---  E(3,2)
!%
!% SW(1,1) --- S(2,1) --- SE(3,1)

    P = stencil(2,2);
    E = stencil(3,2);
    W = stencil(1,2);
    N = stencil(2,3);
    S = stencil(2,1);

    NE = stencil(3,3);
    NW = stencil(1,3);
    SE = stencil(3,1);
    SW = stencil(1,1);

    ans = ( N-S )/(2.d0*eps);

!%! \partial_y \phi 2nd order accuracy
!%	else if ((part_x==0).and.(part_y==1)) then
!%		FDoperator = (PhiN-PhiS)/(2.0*dy);

end function dy

!&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
function dxx(stencil,eps) result(ans)

	implicit none

! Inlet/Outlet
	real(kind=2) :: ans
	real(kind=2),dimension(1:3,1:3),intent(in) :: stencil
	real(kind=2),intent(in) :: eps

    real(kind=2) :: P,E,W,N,S,NE,NW,SE,SW
    
!% NW(1,3) --- N(2,3) --- NE(3,3)
!%
!%  W(1,2) --- P(2,2) ---  E(3,2)
!%
!% SW(1,1) --- S(2,1) --- SE(3,1)

    P = stencil(2,2);
    E = stencil(3,2);
    W = stencil(1,2);
    N = stencil(2,3);
    S = stencil(2,1);

    NE = stencil(3,3);
    NW = stencil(1,3);
    SE = stencil(3,1);
    SW = stencil(1,1);

    ans = ( E-2.d0*P+W )/(eps**2);
        
!%! \partial_x^2 \phi 2nd order accuracy
!%	else if ((part_x==2).and.(part_y==0)) then
!%		FDoperator = (PhiE-2.0*PhiP+PhiW)/(dx**2);

end function dxx

!&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
function dyy(stencil,eps) result(ans)

	implicit none

! Inlet/Outlet
	real(kind=2) :: ans
	real(kind=2),dimension(1:3,1:3),intent(in) :: stencil
	real(kind=2),intent(in) :: eps

    real(kind=2) :: P,E,W,N,S,NE,NW,SE,SW
    
!% NW(1,3) --- N(2,3) --- NE(3,3)
!%
!%  W(1,2) --- P(2,2) ---  E(3,2)
!%
!% SW(1,1) --- S(2,1) --- SE(3,1)

    P = stencil(2,2);
    E = stencil(3,2);
    W = stencil(1,2);
    N = stencil(2,3);
    S = stencil(2,1);

    NE = stencil(3,3);
    NW = stencil(1,3);
    SE = stencil(3,1);
    SW = stencil(1,1);

    ans = ( N-2*P+S )/(eps**2);

!%! \partial_y^2 \phi 2nd order accuracy
!%	else if ((part_x==0).and.(part_y==2)) then
!%		FDoperator = (PhiN-2.0*PhiP+PhiS)/(dy**2);

end function dyy

!&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
function dxy(stencil,eps) result(ans)

	implicit none

! Inlet/Outlet
	real(kind=2) :: ans
	real(kind=2),dimension(1:3,1:3),intent(in) :: stencil
	real(kind=2),intent(in) :: eps

    real(kind=2) :: P,E,W,N,S,NE,NW,SE,SW
    
!% NW(1,3) --- N(2,3) --- NE(3,3)
!%
!%  W(1,2) --- P(2,2) ---  E(3,2)
!%
!% SW(1,1) --- S(2,1) --- SE(3,1)

    P = stencil(2,2);
    E = stencil(3,2);
    W = stencil(1,2);
    N = stencil(2,3);
    S = stencil(2,1);

    NE = stencil(3,3);
    NW = stencil(1,3);
    SE = stencil(3,1);
    SW = stencil(1,1);

    ans = ( NE+SW-NW-SE )/(4.d0*eps**2);

!%! \partial_x\partial_y \phi 2nd order accuracy
!%	else if ((part_x==1).and.(part_y==1)) then
!%		FDoperator = (PhiNE+PhiSW-PhiNW-PhiSE)/(4.0*dx*dy);

end function dxy

!&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
function dxxy(stencil,eps) result(ans)

	implicit none

! Inlet/Outlet
	real(kind=2) :: ans
	real(kind=2),dimension(1:3,1:3),intent(in) :: stencil
	real(kind=2),intent(in) :: eps

    real(kind=2) :: P,E,W,N,S,NE,NW,SE,SW
    
!% NW(1,3) --- N(2,3) --- NE(3,3)
!%
!%  W(1,2) --- P(2,2) ---  E(3,2)
!%
!% SW(1,1) --- S(2,1) --- SE(3,1)

    P = stencil(2,2);
    E = stencil(3,2);
    W = stencil(1,2);
    N = stencil(2,3);
    S = stencil(2,1);

    NE = stencil(3,3);
    NW = stencil(1,3);
    SE = stencil(3,1);
    SW = stencil(1,1);

    ans = ( NE-SE-2.d0*(N-S)+NW-SW )/(2.d0*(eps**2)*eps);

!%! \partial_x^2\partial_y \phi 2nd order accuracy
!%	else if ((part_x==2).and.(part_y==1)) then
!%		FDoperator = (PhiNE-PhiSE-2.0*(PhiN-PhiS)+PhiNW-PhiSW)/(2.0*(dx**2)*dy);

end function dxxy

!&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
function dxyy(stencil,eps) result(ans)

	implicit none

! Inlet/Outlet
	real(kind=2) :: ans
	real(kind=2),dimension(1:3,1:3),intent(in) :: stencil
	real(kind=2),intent(in) :: eps

    real(kind=2) :: P,E,W,N,S,NE,NW,SE,SW
    
!% NW(1,3) --- N(2,3) --- NE(3,3)
!%
!%  W(1,2) --- P(2,2) ---  E(3,2)
!%
!% SW(1,1) --- S(2,1) --- SE(3,1)

    P = stencil(2,2);
    E = stencil(3,2);
    W = stencil(1,2);
    N = stencil(2,3);
    S = stencil(2,1);

    NE = stencil(3,3);
    NW = stencil(1,3);
    SE = stencil(3,1);
    SW = stencil(1,1);

    ans = ( NE-NW-2.d0*(E-W)+SE-SW )/(2.d0*(eps**2)*eps);

!%! \partial_x\partial_y^2 \phi 2nd order accuracy
!%	else if ((part_x==1).and.(part_y==2)) then
!%		FDoperator = (PhiNE-PhiNW-2.0*(PhiE-PhiW)+PhiSE-PhiSW)/(2.0*(dx**2)*dy);

end function dxyy

!&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
function dxxyy(stencil,eps) result(ans)

	implicit none

! Inlet/Outlet
	real(kind=2) :: ans
	real(kind=2),dimension(1:3,1:3),intent(in) :: stencil
	real(kind=2),intent(in) :: eps

    real(kind=2) :: P,E,W,N,S,NE,NW,SE,SW,SumI,SumII
    
!% NW(1,3) --- N(2,3) --- NE(3,3)
!%
!%  W(1,2) --- P(2,2) ---  E(3,2)
!%
!% SW(1,1) --- S(2,1) --- SE(3,1)

    P = stencil(2,2);
    E = stencil(3,2);
    W = stencil(1,2);
    N = stencil(2,3);
    S = stencil(2,1);

    NE = stencil(3,3);
    NW = stencil(1,3);
    SE = stencil(3,1);
    SW = stencil(1,1);

    SumI  = E+N+W+S;
    SumII = NE+NW+SW+SE;
    ans = ( SumII-2.d0*SumI+4.d0*P )/((eps**2)*(eps**2));

!%! \partial_x^2\partial_y^2 \phi 2nd order accuracy
!%	else if ((part_x==2).and.(part_y==2)) then
!%		SumI  = PhiE+PhiN+PhiW+PhiS;
!%		SumII = PhiNE+PhiNW+PhiSW+PhiSE;
!%		FDoperator = (SumII-2.0*SumI+4.0*PhiP)/((dx**2)*(dy**2));

end function dxxyy

!&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
function minmod(a,b,c) result(ans)

	implicit none

! Inlet/Outlet
	real(kind=2) :: ans
	real(kind=2),intent(in) :: a,b,c
	
	real(kind=2),dimension(3) :: vector
	
	vector = (/ a,b,c /)

    if ( (a>0).and.(b>0).and.(c>0) ) then
        ans = minval(vector);
    elseif ( (a<0).and.(b<0).and.(c<0) ) then
        ans = maxval(vector);
    else
        ans = 0.d0;
    endif
    
end function minmod

!&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
function mmdx(stencil,eps,theta) result(ans)

	implicit none

! Inlet/Outlet
	real(kind=2) :: ans
	real(kind=2),dimension(1:3,1:3),intent(in) :: stencil
	real(kind=2),intent(in) :: eps,theta

! Internal Functions
	real(kind=2)    :: minmod

    real(kind=2) :: P,E,W,N,S,NE,NW,SE,SW
    
!% NW(1,3) --- N(2,3) --- NE(3,3)
!%
!%  W(1,2) --- P(2,2) ---  E(3,2)
!%
!% SW(1,1) --- S(2,1) --- SE(3,1)
    
    P = stencil(2,2);
    E = stencil(3,2);
    W = stencil(1,2);
    N = stencil(2,3);
    S = stencil(2,1);

    NE = stencil(3,3);
    NW = stencil(1,3);
    SE = stencil(3,1);
    SW = stencil(1,1);
    
    ans = minmod( theta*(E-P)/eps , (E-W)/(2.d0*eps) , theta*(P-W)/eps );

!% ! \partial_x \phi 2nd order accuracy
!% 	else if ((part_x==1).and.(part_y==0)) then
!% 		FDoperator = (PhiE-PhiW)/(2.0*dx);

end function mmdx

!&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
function mmdy(stencil,eps,theta) result(ans)

	implicit none

! Inlet/Outlet
	real(kind=2) :: ans
	real(kind=2),dimension(1:3,1:3),intent(in) :: stencil
	real(kind=2),intent(in) :: eps,theta

! Internal Functions
	real(kind=2)    :: minmod
	
    real(kind=2) :: P,E,W,N,S,NE,NW,SE,SW
    
!% NW(1,3) --- N(2,3) --- NE(3,3)
!%
!%  W(1,2) --- P(2,2) ---  E(3,2)
!%
!% SW(1,1) --- S(2,1) --- SE(3,1)

    P = stencil(2,2);
    E = stencil(3,2);
    W = stencil(1,2);
    N = stencil(2,3);
    S = stencil(2,1);

    NE = stencil(3,3);
    NW = stencil(1,3);
    SE = stencil(3,1);
    SW = stencil(1,1);

    ans = minmod( theta*(N-P)/eps , (N-S)/(2.d0*eps) , theta*(P-S)/eps );    

!%! \partial_y \phi 2nd order accuracy
!%	else if ((part_x==0).and.(part_y==1)) then
!%		FDoperator = (PhiN-PhiS)/(2.0*dy);

end function mmdy